package be.iminds.ilabt.jfed.testing.tests.highlevel;

import be.iminds.ilabt.jfed.JFedWebApiClientModule;
import be.iminds.ilabt.jfed.JFedWebApiClientTestModule;
import be.iminds.ilabt.jfed.espec.bundle.BundleFetcher;
import be.iminds.ilabt.jfed.espec.bundle.ESpecBundle;
import be.iminds.ilabt.jfed.espec.filefetcher.FileFetcher;
import be.iminds.ilabt.jfed.espec.model.*;
import be.iminds.ilabt.jfed.espec.parser.ExperimentSpecificationParser;
import be.iminds.ilabt.jfed.espec.util.ESpecLogListener;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentController;
import be.iminds.ilabt.jfed.experiment.ExperimentControllerFactory;
import be.iminds.ilabt.jfed.experiment.ExperimentState;
import be.iminds.ilabt.jfed.experimenter_gui.config.JFedExperimenterGuiConfigProvider;
import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfigImpl;
import be.iminds.ilabt.jfed.experimenter_gui.config.UserInfoProvider;
import be.iminds.ilabt.jfed.experimenter_gui.util.ConnectivityDetector;
import be.iminds.ilabt.jfed.experimenter_gui.util.GuiConfigRSpecGenerator;
import be.iminds.ilabt.jfed.experimenter_gui.util.TestbedNodesMapFetcher;
import be.iminds.ilabt.jfed.fedmon.webapi.client.FedmonWebApiClient;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.git.GitFetcher;
import be.iminds.ilabt.jfed.git.GitFetcherCommand;
import be.iminds.ilabt.jfed.highlevel.HighLevelModule;
import be.iminds.ilabt.jfed.highlevel.LowLevelModule;
import be.iminds.ilabt.jfed.highlevel.SfaOnlyExperimentModule;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticAggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticUserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.preferences.CorePreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedCorePreferences;
import be.iminds.ilabt.jfed.rspec.AdvertisementBasedRspecGenerator;
import be.iminds.ilabt.jfed.rspec.RSpecGenerator;
import be.iminds.ilabt.jfed.rspec.RSpecGeneratorFactory;
import be.iminds.ilabt.jfed.rspec.rspec_source.AdvertisementRspecSource;
import be.iminds.ilabt.jfed.testing.base.*;
import be.iminds.ilabt.jfed.testing.shared.CommonAMTest;
import be.iminds.ilabt.jfed.testing.shared.NodeLoginTestStep;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.UserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.testing.shared.Proxy;
import be.iminds.ilabt.jfed.testing.shared.ProxyInfoGenerator;
import be.iminds.ilabt.jfed.util.*;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.inject.*;
import com.google.inject.name.Names;
import com.google.inject.util.Modules;
import io.dropwizard.jackson.Jackson;
import org.apache.commons.compress.archivers.ArchiveEntry;
import org.apache.commons.compress.archivers.ArchiveException;
import org.apache.commons.compress.archivers.ArchiveInputStream;
import org.apache.commons.compress.archivers.ArchiveStreamFactory;
import org.apache.commons.compress.compressors.CompressorException;
import org.apache.commons.compress.compressors.CompressorInputStream;
import org.apache.commons.compress.compressors.CompressorStreamFactory;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.validation.constraints.NotNull;
import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.security.NoSuchAlgorithmException;
import java.time.Duration;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

/**
 * GuiLogicTest
 */
public class GuiLogicTest extends ApiTest<GuiLogicTestConfig> {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(GuiLogicTest.class);

    @Nonnull private final Server userAuthorityServer;
    @Nonnull private final TestbedInfoSource testbedInfoSource;
    @Nonnull private final AuthorityFinder authorityFinder;
    @Nonnull private final RspecFactory rspecFactory;

    //    @Nonnull private final Injector originalInjector;
    @Nonnull private final JFedCorePreferences configJFedPreferences;
    @Nonnull private final Injector configInjector;

    @Inject
    public GuiLogicTest(@Nonnull Injector originalInjector,
                        @Nonnull Logger logger,
                        @Nonnull RspecFactory rspecFactory,
                        @Nonnull TargetAuthority testedAuthority,
                        @Nonnull GeniUserProvider geniUserProvider,
                        @Nonnull GuiLogicTestConfig testConfig) {
        super(logger, testedAuthority, geniUserProvider,
                testConfig);
        this.rspecFactory = rspecFactory;

//        this.originalInjector = originalInjector;
        this.configJFedPreferences = setupPreferences(originalInjector);
        this.configInjector = createInjector(configJFedPreferences);

        this.testbedInfoSource = configInjector.getInstance(TestbedInfoSource.class);
        this.authorityFinder = configInjector.getInstance(AuthorityFinder.class);

        if (user == null)
            throw new IllegalStateException("User is null");

        Server uas = user.getUserAuthorityServer();
        if (uas == null)
            throw new IllegalStateException("User has no Authority Server");

        userAuthorityServer = uas;
    }

    private static final ObjectMapper MAPPER = Jackson.newObjectMapper();
    private static final ApiTestMetaData metadata = new ApiTestMetaData<GuiLogicTestConfig>() {
        @Nonnull
        @Override
        public String getTestDescription() {
            return "A test using the jFed GUI Editor logic. This will reserve resources, test login, and delete them again.";
        }

        @Nonnull
        @Override
        public Class<GuiLogicTestConfig> getConfigClass() {
            return GuiLogicTestConfig.class;
        }

        @Nonnull
        @Override
        public GuiLogicTestConfig parseApiTestConfig(@Nonnull String apiTestConfigString) {
            try {
                return MAPPER.readValue(apiTestConfigString, GuiLogicTestConfig.class);
            } catch (IOException e) {
                throw new RuntimeException("Invalid Test config", e);
            }
        }

        @Nonnull
        @Override
        public ApiTestConfigEditor<GuiLogicTestConfig> getApiTestConfigEditor() {
            return new JsonTestConfigEditor(GuiLogicTestConfig.class);
        }

    };
    public static ApiTestMetaData getMetaData() {
        return metadata;
    }

    NodeLoginTestStep nodeLoginTestStep;




    public AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory getAutomaticUserAndSliceApiWrapperFactory() {
        return configInjector.getInstance(AutomaticUserAndSliceApiWrapper.BasicAutomaticUserAndSliceApiWrapperFactory.class);
//        return new AutomaticUserAndSliceApiWrapper.BasicAutomaticUserAndSliceApiWrapperFactory(
//                logger, geniUserProvider, connectionProvider, getJFedPreferences());
    }

    private JFedCorePreferences setupPreferences(Injector injector) {
        //setup proxy correctly by changing original JFedPreferences
        JFedPreferences basePrefs = injector.getInstance(JFedPreferences.class);
        JFedCorePreferences newJFedPreferences = basePrefs instanceof JFedCorePreferences ?
                ((JFedCorePreferences) basePrefs).makeNoSaveCopy() :
                null;
        if (newJFedPreferences != null) {
            if (getTestConfig().getAmConnectionProxy().getMode().equals(Proxy.ProxyMode.MANUAL))
                throw new RuntimeException("Option amConnectionProxy.mode == "+Proxy.ProxyMode.MANUAL+" is not supported");
            if (getTestConfig().getNodeLoginTest().getProxy().getMode().equals(Proxy.ProxyMode.MANUAL))
                throw new RuntimeException("Option nodeLoginTest.proxy.mode == "+Proxy.ProxyMode.MANUAL+" is not supported");

            newJFedPreferences.setString(
                    CorePreferenceKey.PREF_SSHPROXY_USE_FOR_JFED,
                    getTestConfig().getAmConnectionProxy().getMode().equals(Proxy.ProxyMode.USER_AUTHORITY) ?
                            "ALWAYS" : "NEVER"
            );
            LOG.debug("Set usage of \"call\" proxy to: "+
                    newJFedPreferences.getString(CorePreferenceKey.PREF_SSHPROXY_USE_FOR_JFED));
            newJFedPreferences.setString(
                    CorePreferenceKey.PREF_SSHPROXY_USE_FOR_SSH,
                    getTestConfig().getNodeLoginTest().getProxy().getMode().equals(Proxy.ProxyMode.USER_AUTHORITY) ?
                            "ALWAYS" : "NEVER"
            );
            LOG.debug("Set usage of SSH proxy to: "+
                    newJFedPreferences.getString(CorePreferenceKey.PREF_SSHPROXY_USE_FOR_SSH));
        }
        return newJFedPreferences;
    }


    private Injector createInjector(JFedCorePreferences configPreferences) {
        //just create an injector specifically for us: Do not share this stuff with any other tests that might be around.
        List<Module> minimalModules = new ArrayList<>();
        minimalModules.add(new LowLevelModule());
        minimalModules.add(new HighLevelModule());
        minimalModules.add(new JFedWebApiClientTestModule());
        minimalModules.add(new SfaOnlyExperimentModule());
        minimalModules.add(binder -> binder.bind(GeniUserProvider.class).toInstance(geniUserProvider));
        minimalModules.add(binder -> binder.bind(JFedPreferences.class).toInstance(configPreferences));
        minimalModules.add(binder -> binder.bind(JFedCorePreferences.class).toInstance(configPreferences));

        //make sure correct logger is used
        AbstractModule loggerOverride = new AbstractModule() {
            @Override
            protected void configure() {
                bind(Logger.class).toInstance(getLogger());
                if (getLogger() instanceof JavaFXLogger) {
                    bind(JavaFXLogger.class).toInstance((JavaFXLogger) getLogger());
                } else {
                    //if it's not, that's a problem.
                    // You might run into problems with non-singleton logger later on, which results in calls not being logged.
                    throw new RuntimeException("Expected getLogger() to be a JavaFXLogger.");
                }
            }
        };

        Injector resInjector = Guice.createInjector(Modules.override(minimalModules).with(loggerOverride));

        assertNotNull(resInjector);
        LOG.debug("Created injector");

        Logger injectorLogger = resInjector.getInstance(Logger.class);
        //noinspection ObjectEquality
        if (getLogger() != injectorLogger) {
            errorFatal("Invalid logger in injector");
        } else {
            LOG.debug("Logger OK");
        }

        JFedPreferences injectorPreferences = resInjector.getInstance(JFedPreferences.class);
        //noinspection ObjectEquality
        if (configPreferences != injectorPreferences) {
            errorFatal("Invalid JFedPreferences in injector");
        } else {
            LOG.debug("JFedPreferences OK");
        }

        return resInjector;
    }

    @Override
    public void setUp() throws Exception {
        assertNotNull(geniUserProvider, "No user specified for test");
        assertNotNull(user, "No user specified for test");
        assertNotNull(getTestedAuthority(), "No target server specified for test");
        assertNotNull(getTestConfig(), "no test config");
        assertTrue(getTestConfig().isValid(), "Test config is not valid");
        assertNotEmpty(getTestConfig().getResources(), "Configuration error: no Resource config");
        AtomicInteger index = new AtomicInteger(0);
        resourceInfos = getTestConfig().getResources().stream()
                .filter(Objects::nonNull)
                .map(r -> new ResourceInfo(index.getAndIncrement(), r))
                .collect(Collectors.toList());
        assertEquals(resourceInfos.size(), getTestConfig().getResources().size(), "Configuration error: some of the Resources in the config where null");

        assertNotEmpty(resourceInfos, "bug");

        this.nodeLoginTestStep =
                new NodeLoginTestStep(this, getTestConfig().generateNodeLoginTestStepConfig(geniUserProvider));

        //fail early
        for (ResourceInfo resourceInfo : resourceInfos) {
            assertTrue(resourceInfo.config.getProvision().isEnabled(),
                    "A disabled provision step is currently not supported.");
        }

        setupAnsible();
    }

    /** Info on a Resource that is shared between test methods. */
    private class ResourceInfo {
        private final int index;

        @NotNull private final GuiLogicTestConfig.Resource config;

        private GeniUrn sliceUrn;

        private String requestRspec;
        private int maxNodeNameLength = 5; //TODO: automatically get from rspec
        private List<Server> serversInRspec;
        private RequestRspecSource requestRspecSource;
        public String sliceName;

        private ESpecBundle eSpecBundle;
        private ExperimentSpecification eSpec;

        private ESpecLogToTest eSpecLogToTest;

        public ResourceInfo(int index, @NotNull GuiLogicTestConfig.Resource config) {
            this.index = index;
            this.config = config;
        }

        private String project;
        private Experiment experiment;
        private boolean notEnoughResourcesDetected = false;
        private ExperimentController experimentController;

        private boolean waitForExperimentIsOver = false;
        private boolean deleted = false;

        String manifestRspecString;
        boolean validManifest;

        public boolean isWaitForExperimentOver() {
            return waitForExperimentIsOver;
        }

        public boolean isDeleted() {
            return deleted;
        }

        public boolean isNotEnoughResourcesDetected() {
            return notEnoughResourcesDetected;
        }
    }

    private List<ResourceInfo> resourceInfos;


    private JFedGuiConfig cachedGuiConfig = null;
    private JFedGuiConfig getGuiConfig() {
        if (cachedGuiConfig != null)
            return cachedGuiConfig;

        JFedPreferences jFedPreferences = configInjector.getInstance(JFedPreferences.class);
        ConnectivityDetector connectivityDetector = configInjector.getInstance(ConnectivityDetector.class);
        UserInfoProvider userInfoProvider = configInjector.getInstance(UserInfoProvider.class);
        URL webApiUrl =
                configInjector.getInstance(Key.get(URL.class, Names.named(JFedWebApiClientModule.NAMED__NO_CLIENT_AUTH_WEBAPI_URL)));
        FedmonWebApiClient fedmonWebApiClient = configInjector.getInstance(FedmonWebApiClient.class);

        if (geniUserProvider == null) {
            throw new NullPointerException("geniUserProvider may not be null");
        }

        JFedExperimenterGuiConfigProvider jFedExperimenterGuiConfigProvider =
                new JFedExperimenterGuiConfigProvider(
                        geniUserProvider,
                        jFedPreferences,
                        connectivityDetector,
                        userInfoProvider,
                        webApiUrl,
                        logger,
                        false);
        TestbedNodesMapFetcher testbedNodesMapFetcher =
                new TestbedNodesMapFetcher(connectivityDetector);
        cachedGuiConfig = new JFedGuiConfigImpl(
                testbedInfoSource,
                authorityFinder,
                jFedExperimenterGuiConfigProvider,
                testbedNodesMapFetcher,
                fedmonWebApiClient);
        return cachedGuiConfig;
    }

    private RSpecGeneratorFactory getRSpecGeneratorFactory() {
        return new RSpecGeneratorFactory() {
            @Nonnull
            @Override
            public RSpecGenerator create() {
                return new GuiConfigRSpecGenerator(
                        getGuiConfig(),
                        authorityFinder,
                        testbedInfoSource);
            }
        };
    }


    private AdvertisementRspecSource advertisementRspecSource;
    @Test(groups = {"init"})
    public void fetchAdvertisementRspec() throws JFedException {
        boolean needAdvertisementRspec = false;
        for (ResourceInfo resourceInfo : resourceInfos) {
            GuiLogicTestConfig.RequestRSpec rspecConfig = resourceInfo.config.getRequestRSpec();
            if (rspecConfig != null) {
                if (rspecConfig.getSource() == GuiLogicTestConfig.RequestRSpecSource.GENERATE_USING_ADVERTISEMENT) {
                    needAdvertisementRspec = true;
                    break;
                }
            }
        }
        if (!needAdvertisementRspec) {
            note("Advertisement RSpec not needed. Will do nothing here.");
            return;
        }

        if (userAndSliceApiWrapper == null) {
            userAndSliceApiWrapper =
                    this.getAutomaticUserAndSliceApiWrapperFactory().create();
        }
        assertNotNull (user, "internal error");
        GeniUrn userUrn = user.getUserUrn();
        List<AnyCredential> userCredentials =
                userAndSliceApiWrapper.getUserCredentials(logger, userUrn);

        AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory amWrapperFact =
                configInjector.getInstance(AutomaticAggregateManagerWrapper.BasicAutomaticAggregateManagerWrapperFactory.class);
        AggregateManagerWrapper amWrapper = amWrapperFact.create(testedAuthority.getServerToConnect());
        String advertismentRspec = amWrapper.listResources(userCredentials, true/*available*/);
        if (advertismentRspec == null) {
            errorFatal("ListResources call did not return an advertisement RSpec");
            return;
        }

        this.advertisementRspecSource = new AdvertisementRspecSource(advertismentRspec, ModelRspecType.BASIC);
    }


    @Test(hardDepends = {"fetchAdvertisementRspec"}, groups = {"init"})
    public void getRspec() throws GeniUrn.GeniUrnParseException,
            ESpecBundle.ESpecBundleInitException,
            InterruptedException,
            IOException,
            ExperimentSpecificationParser.ExperimentSpecificationParseException {
        int resourceIndex = 0;
        for (ResourceInfo resourceInfo : resourceInfos) {
            GuiLogicTestConfig.RequestRSpec rspecConfig = resourceInfo.config.getRequestRSpec();
            if (rspecConfig != null) {
                switch (rspecConfig.getSource()) {
                    case GENERATE_USING_ADVERTISEMENT: {
                        GuiLogicTestConfig.GenerateUsingAdvertisementRequestRSpec specRspecConfig =
                                (GuiLogicTestConfig.GenerateUsingAdvertisementRequestRSpec)rspecConfig;
                        assertNotNull(this.advertisementRspecSource, "internal error");
                        assertNotNull(specRspecConfig.getConfig(), "internal error: check not done yet");
                        AdvertisementBasedRspecGenerator gen =
                                new AdvertisementBasedRspecGenerator(
                                        rspecFactory,
                                        specRspecConfig.getConfig(),
                                        this.advertisementRspecSource
                                );
                        resourceInfo.requestRspec = gen.generate().getRspecXmlString();
                        note("Generated Request RSpec using \"GENERATE_USING_ADVERTISEMENT\" method:");
                        note(resourceInfo.requestRspec, true);
                        break;
                    }
                    case GENERATE_USING_JFED: {
                        GuiLogicTestConfig.GenerateUsingJFedRequestRSpec specRspecConfig =
                                (GuiLogicTestConfig.GenerateUsingJFedRequestRSpec)rspecConfig;
                        RSpecGenerator gen = getRSpecGeneratorFactory().create();
                        gen.setAm(getTestedAuthority().getRspecUrn().toString());
                        gen.setNodeCount(specRspecConfig.getNodeCount());
                        gen.setResourceClassIdPrefix(specRspecConfig.getjFedResourceClass());
                        gen.setNodeAnsibleGroup(specRspecConfig.getNodeAnsibleGroup());
                        resourceInfo.requestRspec = gen.generateRequest();
                        note("Generated Request RSpec using \"GENERATE_USING_JFED\" method:");
                        note(resourceInfo.requestRspec, true);
                        break;
                    }
                    case GENERATE_SIMPLE: {
                        GuiLogicTestConfig.GenerateSimpleRequestRSpec specRspecConfig =
                                (GuiLogicTestConfig.GenerateSimpleRequestRSpec)rspecConfig;
                        @Nonnull String exclusive = specRspecConfig.getNodeExclusive() ? "true" : "false";
                        @Nonnull String sliverTypeName = specRspecConfig.getNodeSliverType();
                        int nodeCount = specRspecConfig.getNodeCount();
                        @Nonnull List<String> compIds = specRspecConfig.getNodeComponentUrn();
                        assertTrue(compIds.isEmpty() || nodeCount == compIds.size(), "Config error: Mismatch of componentIds and nodeCount");
                        @Nonnull String compManId = getTestedAuthority().getServerForRspecComponentManager().getDefaultComponentManagerUrn();
                        resourceInfo.requestRspec = "<?xml version='1.0'?>\n" +
                                "<rspec xmlns=\"http://www.geni.net/resources/rspec/3\" type=\"request\"  " +
                                "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" " +
                                "xsi:schemaLocation=\"http://www.geni.net/resources/rspec/3 http://www.geni.net/resources/rspec/3/request.xsd \">\n";

                        for (int i = 0; i < nodeCount; i++) {
                            String ansibleGroup = i < specRspecConfig.getNodeAnsibleGroup().size() ?
                                    specRspecConfig.getNodeAnsibleGroup().get(i) :
                                    (specRspecConfig.getNodeAnsibleGroup().isEmpty() ?
                                            null :
                                            specRspecConfig.getNodeAnsibleGroup().get(specRspecConfig.getNodeAnsibleGroup().size()-1)
                                    );

                            @Nonnull String compId = "";
                            if (compIds.size() > i)
                                compId = " component_id=\"" + compIds.get(i) + "\"";
                            resourceInfo.requestRspec +=
                                    "  <node client_id=\"n" + i + "\" exclusive=\"" + exclusive + "\" component_manager_id=\"" + compManId + "\"" + compId + ">\n" +
                                            "    <sliver_type name=\"" + sliverTypeName + "\"/>\n" +
                                            (ansibleGroup == null ?
                                                    "" :
                                                    "    <ansible_group xmlns=\"http://jfed.iminds.be/rspec/ext/jfed/1\" name=\""+ansibleGroup+"\"/>\n"
                                            )+
                                            "  </node>\n";
                        }

                        resourceInfo.requestRspec += "</rspec>";

                        note("Generated Request RSpec using \"simple\" method:");
                        note(resourceInfo.requestRspec, true);
                        break;
                    }
                    case PROVIDE_URL: {
                        GuiLogicTestConfig.ProvidedRequestRSpec specRspecConfig =
                                (GuiLogicTestConfig.ProvidedRequestRSpec)rspecConfig;
                        String rspecUrl = specRspecConfig.getProvidedContentSource();
                        assertNotNull(rspecUrl, "Config error: no RSpec URL provided");

                        URL url;
                        try {
                            url = new URL(rspecUrl);
                        } catch (MalformedURLException e) {
                            throw new RuntimeException("Invalid URL specified for RSpec: \"" + rspecUrl + "\"", e);
                        }
                        try (InputStream is = url.openStream()) {
                            String rspec = IOUtils.streamToString(is, "UTF-8");
                            if (rspec.trim().isEmpty()) {
                                warn("Empty Rspec read from URL \"" + rspecUrl + "\"");
                            }

                            String head = TextUtil.abbreviate(rspec, 200);
                            note("Downloaded Rspec from \"" + rspecUrl + "\". Rspec size: " + rspec.length() + " characters.\nHead: " + head, true/*formatAsCode*/);
                            resourceInfo.requestRspec = rspec;
                        } catch (IOException e) {
                            throw new RuntimeException("Failed to fetch RSpec at \"" + rspecUrl + "\"", e);
                        }
                        break;
                    }
                    case PROVIDE_CONTENT: {
                        GuiLogicTestConfig.ProvidedRequestRSpec specRspecConfig =
                                (GuiLogicTestConfig.ProvidedRequestRSpec)rspecConfig;
                        resourceInfo.requestRspec = specRspecConfig.getProvidedContentSource();
                        assertNotNull(resourceInfo.requestRspec, "Config error: no RSpec provided");

                        String head = TextUtil.abbreviate(resourceInfo.requestRspec, 200);
                        note("Using Request Rspec from config. " +
                                "Size: " + resourceInfo.requestRspec.length() + " characters.\n" +
                                "Head: " + head, true/*formatAsCode*/);

                        break;
                    }
                    default:
                        throw new RuntimeException("RequestRSpec source " + resourceInfo.config.getRequestRSpec().getSource() + " not yet implemented");
                }

                //noinspection ConstantConditions
                if (rspecConfig.getSource().isProvide() &&
                        ((GuiLogicTestConfig.ProvidedRequestRSpec)rspecConfig)
                                .getBindTestedServerToUnboundNodes()) {
                    String origRequestRspec = resourceInfo.requestRspec;
                    String targetServerCM = testedAuthority.getServerForRspecComponentManager().getDefaultComponentManagerUrn();
                    resourceInfo.requestRspec = CommonAMTest.bindUnboundTo(this, resourceInfo.requestRspec, targetServerCM);
                    if (!origRequestRspec.equals(resourceInfo.requestRspec)) {
                        note("Unbound nodes were bound to " + targetServerCM);
                    }
                }

                currentTestResult.addExtraResult(
                        new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                                .setKey("rspec" + resourceIndex)
                                .setValue(resourceInfo.requestRspec)
                                .setLarge(true)
                                .setHttpMediaTypeToXml()
                );

                resourceInfo.requestRspecSource = new RequestRspecSource(resourceInfo.requestRspec, ModelRspecType.BASIC);
                resourceInfo.serversInRspec = resourceInfo.requestRspecSource.getAllComponentManagerUrns().stream()
                        .map(serverUrn ->
                                authorityFinder.findByUrn(
                                        serverUrn,
                                        AuthorityFinder.Purpose.REQUEST_RSPEC)
                        )
                        .collect(Collectors.toList());

                resourceInfo.maxNodeNameLength = resourceInfo.requestRspecSource.getBasicNodeInfo().stream()
                        .map(BasicStringRspec.BasicNodeInfo::getClientId)
                        .filter(Objects::nonNull)
                        .mapToInt(String::length)
                        .max()
                        .orElse(5); //uses 5 as fallback when no nodes with client_id found
            }

            ESpecTestConfig.ESpec eConfig = resourceInfo.config.getESpec();
            if (eConfig != null) {
                assertNotNull(eConfig.getProvidedContentSource());

                switch (eConfig.getSource()) {
                    case PROVIDE_ARCHIVE_FILE: {
                        resourceInfo.eSpecBundle = BundleFetcher.fetchArchiveFile(eConfig.getProvidedContentSource());
                        break;
                    }
                    case PROVIDE_DIR: {
                        resourceInfo.eSpecBundle = BundleFetcher.fetchDir(eConfig.getProvidedContentSource());
                        break;
                    }
                    case PROVIDE_ARCHIVE_URL: {
                        resourceInfo.eSpecBundle = BundleFetcher.fetchArchiveUrl(eConfig.getProvidedContentSource());
                        break;
                    }
                    case PROVIDE_GIT_REPO_DIR: {
                        resourceInfo.eSpecBundle = BundleFetcher.fetchGitRepoDir(eConfig.getProvidedContentSource());
                        break;
                    }
                    default: throw new IllegalStateException("Unsupported espec.source: "+eConfig.getSource().name());
                }
                assertNotNull(resourceInfo.eSpecBundle);

                currentTestResult.addExtraResult(
                        new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                                .setKey(ESpecBundle.ESPEC_DESCRIPTION_FILENAME+
                                        (resourceInfos.size() > 0 ? ("-"+resourceIndex) : ""))
                                .setValue(resourceInfo.eSpecBundle.getExperimentSpecificationYml())
                                .setLarge(true)
                                .setHttpMediaTypeToXml()
                );

                ExperimentSpecificationParser parser = new ExperimentSpecificationParser();
                resourceInfo.eSpec = parser.parse(resourceInfo.eSpecBundle.getExperimentSpecificationYml());

                String requestRspec =
                        new String(new FileFetcher(
                                resourceInfo.eSpec.getRspecs().get(0).getSource(),
                                resourceInfo.eSpecBundle,
                                getRSpecGeneratorFactory()).fetchBytes(),
                                StandardCharsets.UTF_8);
                RequestRspecSource requestRspecSource = new RequestRspecSource(requestRspec, ModelRspecType.BASIC);
                resourceInfo.serversInRspec = requestRspecSource.getAllComponentManagerUrns().stream()
                        .map(serverUrn ->
                                authorityFinder.findByUrn(
                                        serverUrn,
                                        AuthorityFinder.Purpose.REQUEST_RSPEC)
                        )
                        .collect(Collectors.toList());

                resourceInfo.maxNodeNameLength = requestRspecSource.getBasicNodeInfo().stream()
                        .map(BasicStringRspec.BasicNodeInfo::getClientId)
                        .filter(Objects::nonNull)
                        .mapToInt(String::length)
                        .max()
                        .orElse(5); //uses 5 as fallback when no nodes with client_id found

                currentTestResult.addExtraResult(
                        new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                                .setKey("rspec")
                                .setValue(requestRspec)
                                .setLarge(true)
                                .setHttpMediaTypeToXml()
                );
            }

            resourceIndex++;
        }
    }

    private UserAndSliceApiWrapper userAndSliceApiWrapper;
    @Test(hardDepends = {"getRspec"}, groups = {"init"})
    public void findProject() {
        assertNotNull(user);
        if (userAndSliceApiWrapper == null) {
            userAndSliceApiWrapper =
                    this.getAutomaticUserAndSliceApiWrapperFactory().create();
        }

        for (ResourceInfo resourceInfo : resourceInfos) {
            switch (resourceInfo.config.getSlice().getProjectSource()) {
                case NONE: {
                    resourceInfo.project = null;
                    break;
                }
                case PROVIDED: {
                    resourceInfo.project = resourceInfo.config.getSlice().getProject();
                    break;
                }
                case AUTO: {
                    resourceInfo.project = CommonAMTest.findProject(
                            this,
                            this.getLogger(),
                            resourceInfo.project,
                            user,
                            user.getUserUrn(),
                            userAndSliceApiWrapper,
                            true);
                    break;
                }
            }

            if (resourceInfo.project != null) {
                note("Will use project: \"" + resourceInfo.project + "\"");
            } else {
                assertFalse(resourceInfo.config.getSlice().getFailIfNoProject(), "Empty project is not allowed by config");
                note("Will NOT use a project");
            }
        }
    }

    @Test(hardDepends = {"getRspec", "findProject"}, groups = {"init"})
    public void createExperiment() {
        assertNotNull(user);
        for (ResourceInfo resourceInfo : resourceInfos) {
            resourceInfo.sliceName = CommonAMTest.createSliceName(
                    this,
                    userAuthorityServer,
                    ApiInfo.getDefaultComponentManagerForEach(resourceInfo.serversInRspec),
                    "g",
                    resourceInfo.maxNodeNameLength,
                    null);

            note("Will create slice \"" + resourceInfo.sliceName + "\"");

            Instant requestedEndTime = Instant.now().plus(resourceInfo.config.getSlice().getExpireTimeMin(), ChronoUnit.MINUTES);
            note("Created slice (and slivers) will expire at " + requestedEndTime + "");

//        List<String> sshKeys = Collections.singletonList(KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(user.getPublicKey()));
            List<String> sshKeys = Collections.singletonList(nodeLoginTestStep.getSshKeyHelper().getSshPublicKeyString());
            UserSpec userSpec = new UserSpec(user.getUserUrnString(), sshKeys);
            List<UserSpec> userSpecs = Collections.singletonList(userSpec);

            if (resourceInfo.eSpec == null) {
                resourceInfo.experiment = new Experiment(
                        resourceInfo.sliceName,
                        resourceInfo.project,
                        resourceInfo.requestRspecSource,
                        null,
                        requestedEndTime,
                        userSpecs,
                        false);
            } else {
                resourceInfo.experiment = new Experiment(
                        resourceInfo.sliceName,
                        resourceInfo.project,
                        resourceInfo.eSpecBundle,
                        null,
                        requestedEndTime,
                        userSpecs,
                        false,
                        getRSpecGeneratorFactory());

                resourceInfo.eSpecLogToTest = new ESpecLogToTest(resourceInfo.index+"");
                resourceInfo.experiment.addESpecLogListener(resourceInfo.eSpecLogToTest);
            }

            if (resourceInfo.config.getWaitForReady().isEnabled()) {
                resourceInfo.experiment.setMaxWaitUntilReady(
                        Duration.ofMinutes(resourceInfo.config.getWaitForReady().getMaxTimeMin())
                );

                //The automated tester does not need to check each 10 seconds like the GUI
                resourceInfo.experiment.setCheckReadyInterval(Duration.ofSeconds(30));

                if (resourceInfo.experiment.getMaxWaitUntilReady() != null) {
                    note("Configured experiment for maximum wait time of " +
                            resourceInfo.experiment.getMaxWaitUntilReady().toMinutes() + " minutes");
                }
            }
        }
    }

    @Test(hardDepends = {"createExperiment"}, groups = {"run"})
    public void runExperiment() {
        for (ResourceInfo resourceInfo : resourceInfos) {
            assertTrue(resourceInfo.config.getProvision().isEnabled(),
                    "A disabled provision step is currently not supported.");

            resourceInfo.notEnoughResourcesDetected = false;

            final ExperimentControllerFactory experimentControllerFactory =
                    configInjector.getInstance(ExperimentControllerFactory.class);
            assertNotNull(experimentControllerFactory);

            resourceInfo.experimentController =
                    experimentControllerFactory.createExperimentController(resourceInfo.experiment);
            assertNotNull(resourceInfo.experimentController);

//        experimentController.setExperimentTester(experimenterJobFactory::createTestConnectivityJob);
//        experimentController.addListener(bugReportExperimentControllerListener);

            Logger injectorLogger = configInjector.getInstance(Logger.class);
            //noinspection ObjectEquality
            if (getLogger() != injectorLogger) {
                errorFatal("Logger check: Invalid logger in experimentController injector");
            } else {
                note("Logger check: OK");
            }

            resourceInfo.experimentController.start();
            note("Started experimentController. ExperimentState is now " + resourceInfo.experiment.getExperimentState());
        }

        //Now wait.
        //This cannot be in a seperate method: we might miss calls in between the 2 methods otherwise (because experimentController is async).

        long maxTimeToWaitForReadyInMinutes = 0;
        long extraWaitTimeInSeconds = 0;
        for (ResourceInfo resourceInfo : resourceInfos) {
            if (!resourceInfo.config.getWaitForReady().isEnabled()) {
                //don't take this wait and extra time into account

                //consider the wait over
                resourceInfo.waitForExperimentIsOver = true;
                continue;
            }
            if (resourceInfo.config.getWaitForReady().getMaxTimeMin() > maxTimeToWaitForReadyInMinutes) {
                maxTimeToWaitForReadyInMinutes = resourceInfo.config.getWaitForReady().getMaxTimeMin();
            }
            if (resourceInfo.config.getWaitForReady().getExtraWaitTimeSeconds() > extraWaitTimeInSeconds) {
                extraWaitTimeInSeconds = resourceInfo.config.getWaitForReady().getExtraWaitTimeSeconds();
            }
        }

        if (maxTimeToWaitForReadyInMinutes > 0) {
            note("Start waiting (max "+maxTimeToWaitForReadyInMinutes+" minutes)" +
                    " for experiment(s) to become ready at " + (new Date().toString()));
        } else {
            note("Will not wait for experiment(s) to become ready.");
        }

        long now = System.currentTimeMillis();
        long deadline = now + (maxTimeToWaitForReadyInMinutes * 60 * 1000L);
        boolean allDone = false;
        boolean sawFailure = false;
        while (now < deadline && !allDone && !sawFailure) {
            for (ResourceInfo resourceInfo : resourceInfos) {
                if (!resourceInfo.waitForExperimentIsOver) {
                    assertTrue(resourceInfo.config.getWaitForReady().isEnabled(), "bug in runExperiment step code.");
                    note("Current ExperimentState=" + resourceInfo.experiment.getExperimentState());
                    note("Checking if experiment is ready.");

                    if (resourceInfo.experiment.getSliceUrn() != null) {
                        resourceInfo.sliceUrn = resourceInfo.experiment.getSliceUrn();
                    }

                    if (resourceInfo.experiment.getExperimentState() == ExperimentState.READY) {
                        note("Ready, because ExperimentState is READY");
                        resourceInfo.waitForExperimentIsOver = true;
                    }
                    if (resourceInfo.experiment.getAnyPartSawNotEnoughFreeResourcesError()) {
                        resourceInfo.waitForExperimentIsOver = true;
                        resourceInfo.notEnoughResourcesDetected = true;
                        sawFailure = true;
                        note("An experiment failed: not enough free resources");
                        continue;
                    }
                    if (resourceInfo.experiment.getExperimentState().isTerminated()) {
                        resourceInfo.waitForExperimentIsOver = true;
                        sawFailure = true;
                        note("Failure, because ExperimentState is terminated");
                    }
                    if (!resourceInfo.experiment.getExperimentState().isActiveState()) {
                        resourceInfo.waitForExperimentIsOver = true;
                        sawFailure = true;
                        note("Failure, because ExperimentState is not active state");
                    }
                }
            }

            allDone = resourceInfos.stream().map(ResourceInfo::isWaitForExperimentOver).allMatch(b -> b == true);

            if (!allDone && !sawFailure) {
                try {
                    Thread.sleep(30000 /*ms*/);
                } catch (InterruptedException e) { /* Ignore*/ }
            }
            now = System.currentTimeMillis();
        }

        if (maxTimeToWaitForReadyInMinutes > 0) {
            note("Stopped waiting for experiment(s) to become ready at " + (new Date().toString()));
        }

        if (extraWaitTimeInSeconds > 0) {
            note("Waiting "+extraWaitTimeInSeconds+" extra seconds before continuing. (due to resource.waitForReady.extraWaitTimeSeconds option)");
            try {
                Thread.sleep(extraWaitTimeInSeconds * 1000L);
            } catch (InterruptedException e) {
                warn("Extra waiting time was interrupted. This should not happen.", e);
            }
        }

        for (ResourceInfo resourceInfo : resourceInfos) {
            if (resourceInfo.notEnoughResourcesDetected) {
                warn("Test skipped because not enough free resources detected while tying to create the sliver(s)");
                return;
            }
        }

        if (resourceInfos.stream().anyMatch(ResourceInfo::isNotEnoughResourcesDetected)) {
            warn("Test cannot continue: Not enough free resources. ");
            note("This test method will be be in the \"WARNING\" state because of \"Not enough free resources\".");
            if (!sawFailure) {
                warn("(Additional warning: internal state not set to failure despite this.)");
                return;
            }
            return;
        }

        if (sawFailure) {
            errorFatal("One or more experiments have failed.");
        }

        if (!allDone) {
            errorFatal("One or more experiments did not become ready by the deadline.");
        }
    }

    public boolean isValidGeniRspec(String rspec) {
        return new BasicStringRspec(rspec).isValidRspec();
    }


    @Test(hardDepends = {"runExperiment"}, groups = {"login"})
    public void checkManifest() throws JFedException {
        for (ResourceInfo resourceInfo : resourceInfos) {
            if (resourceInfo.notEnoughResourcesDetected) {
                warn("Test skipped because not enough free resources detected while tying to create the sliver(s)");
                return;
            }

            Slice slice = resourceInfo.experiment.getSliceOrNull();
            assertNotNull(slice, "experiment slice should not be null");
            ManifestRspecSource manifestRspecSource = slice.getManifestRspec();
            assertNotNull(manifestRspecSource, "experiment slice manifestRspecSource should not be null");
            resourceInfo.manifestRspecString =  manifestRspecSource.getRspecXmlString();
            assertNotNull(resourceInfo.manifestRspecString, "experiment slice manifestRspecSource getRspecXmlString should not be null");

            boolean validRequest = isValidGeniRspec(resourceInfo.requestRspec);
            resourceInfo.validManifest = isValidGeniRspec(resourceInfo.manifestRspecString);

            if (resourceInfo.requestRspec != null && validRequest && resourceInfo.validManifest) {
                setErrorsNotFatal();
                assertTrue(CommonAMTest.sameNodesInRequestAndManifest(this, resourceInfo.requestRspec, resourceInfo.manifestRspecString));
                setErrorsFatal();
            }

            if (resourceInfo.validManifest) {
                boolean foundNodeLogin = nodeLoginTestStep.parseSshInfoFromGeni3ManifestRspec(resourceInfo.manifestRspecString, null, null);
                if (foundNodeLogin)
                    note("Successfully found node login in manifest");
                else
                    note("Did not find node login info in manifest");

                if (resourceInfo.config.getWaitForReady().getSkipFinalManifestCheck())
                    nodeLoginTestStep.checkManifestCorrectness(resourceInfo.manifestRspecString, testedAuthority);
            }
        }
    }

    @Test(hardDepends = {"checkManifest"}, groups = {"login"})
    public void login() throws NoSuchAlgorithmException, IOException {
        if (!getTestConfig().getNodeLoginTest().isEnabled()) {
            skip("Test skipped because disabled in config");
            return;
        }

        //todo; //TODO use nodes from different RSpecs for node login test
        //todo; //TODO log in to multiple nodes

        assertNotNull(geniUserProvider, "No user specified for test");
        assert nodeLoginTestStep != null;

        for (ResourceInfo resourceInfo : resourceInfos) {
            if (resourceInfo.notEnoughResourcesDetected) {
                warn("Test skipped because not enough free resources detected while tying to create the sliver(s)");
                return;
            }
            if (!resourceInfo.validManifest) {
                skip("Test skipped because at least one manifest is not a geni version 3 RSpec (so this test does not know how to extract node login info)");
                return;
            }
        }

        String sliceName = resourceInfos.get(0).sliceName;

        nodeLoginTestStep.testNodeLogin(false/*only note, no warning*/, geniUserProvider, sliceName);
    }

    private JFedConnection.SshProxyInfo convertProxy(be.iminds.ilabt.jfed.testing.shared.Proxy proxy, ResourceInfo resourceInfo) {
        assertNotNull(geniUserProvider, "No user specified for test");
        return ProxyInfoGenerator.generate(this, proxy, geniUserProvider, resourceInfo.sliceName, nodeLoginTestStep.getSshKeyHelper(), nodeLoginTestStep.getSshUsername());
    }


    @Test(hardDepends = {"checkManifest"}, groups = {"login"})
    public void ansible() throws NoSuchAlgorithmException, IOException {
        if (!getTestConfig().getAnsibleTest().isEnabled()) {
            skip("Test skipped because disabled in config");
            return;
        }

        LOG.debug("Starting ansible test");
        JFedConnection.SshProxyInfo sshProxy = convertProxy(getTestConfig().getAnsibleTest().getProxy(), resourceInfos.get(0));

        assertTrue(sshProxy == null, "ansible does not support an SSH proxy, but proxy configured: " + sshProxy);

        try {
            writeAnsibleFiles(getTestConfig().getAnsibleTest().getDebug());
            LOG.debug("   wrote ansible files to " + ansibleDir);
        } catch (IOException e) {
            errorFatal("Failed to write ansible files", e);
        }
        String output = callAnsible();
        LOG.debug("    got ansible output");

        if (output != null) {
            currentTestResult.addExtraResult(
                    new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                            .setKey("ansible", "output")
                            .setValue(output)
                            .setLarge(true)
                            .setHttpMediaTypeToText());

            List<GuiLogicTestConfig.ExtractSpec> extractSpecs =
                    getTestConfig().getAnsibleTest().getExtract();
            for (GuiLogicTestConfig.ExtractSpec extract : extractSpecs) {
                assert !extract.getDelim().trim().isEmpty();
                Matcher m = Pattern.compile(
                        Pattern.quote(extract.getDelim()) + "(.+?)" + Pattern.quote(extract.getDelim())
                ).matcher(output);
                if (m.find()) {
                    String extracted = m.group(1);
                    currentTestResult.addExtraResult(
                            new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                                    .setKey("extract", extract.getName())
                                    .setValue(extracted)
                                    .setLarge(false));
                    note("Found value to extract with name=\""+extract.getName()+"\" delim=\""+extract.getDelim()+"\"");
                } else {
                    note("Did NOT find value to extract with name=\""+extract.getName()+"\" delim=\""+extract.getDelim()+"\"");
                }
            }
        }

        if (output == null || output.trim().isEmpty())
            fatalError("Ansible did not return any output.");

        note("Ansible output:\n" + output + "\n", true/*formatAsCode*/);

        checkAnsibleOutput(
                output, this,
                null, null, null,
                getTestConfig().getAnsibleTest().getSuccessRegex(),
                getTestConfig().getAnsibleTest().getWarningRegex(),
                getTestConfig().getAnsibleTest().getFailureRegex());
    }

    @Test(softDepends = {"login", "ansible", "runExperiment"},
            hardDepends = {"getRspec", "createExperiment"},
            groups = {"cleanup", "login", "run"})
    public void deleteExperiment() {
        for (ResourceInfo resourceInfo : resourceInfos) {
            if (resourceInfo.config.getCleanup().isEnabled()) {
                note("Will call experimentController.stop();");
                resourceInfo.experimentController.stop();
            } else {
                note("Experiment Cleanup disabled");
                resourceInfo.deleted = true;
            }
        }

        long maxTimeToWaitForDeleteInMinutes = 0;
        for (ResourceInfo resourceInfo : resourceInfos) {
            if (resourceInfo.config.getCleanup().getMaxWaitTimeMinutes() > maxTimeToWaitForDeleteInMinutes) {
                maxTimeToWaitForDeleteInMinutes = resourceInfo.config.getCleanup().getMaxWaitTimeMinutes();
            }
        }

        note("Will wait until experiment is terminated");
        long now = System.currentTimeMillis();
        long deadline = now + (maxTimeToWaitForDeleteInMinutes * 60 * 1000L);
        boolean allDeleted = false;
        while (now < deadline && !allDeleted) {
            for (ResourceInfo resourceInfo : resourceInfos) {
                if (!resourceInfo.deleted) {
                    note("Current ExperimentState=" + resourceInfo.experiment.getExperimentState());
                    if (resourceInfo.experiment.getExperimentState() == ExperimentState.EMPTY ||
                            resourceInfo.experiment.getExperimentState() == ExperimentState.EXPIRED ||
                            resourceInfo.experiment.getExperimentState() == ExperimentState.FAILED) {
                        note("Experiment was terminated");
                        resourceInfo.deleted = true;
                    }
                }
            }

            allDeleted = resourceInfos.stream().map(ResourceInfo::isDeleted).allMatch(b -> b == true);

            if (!allDeleted) {
                try {
                    Thread.sleep(30000 /*ms*/);
                } catch (InterruptedException e) { /* Ignore*/ }
            }
            now = System.currentTimeMillis();
        }

        if (!allDeleted) {
            errorFatal("Experiment(s) did not terminate within " + maxTimeToWaitForDeleteInMinutes + " minutes. Gave up waiting for it.");
        } else {
            note("All experiments deleted");
        }
    }




    private File ansibleDir;
    private File ansiblePlaybookLocalFile;

    private String ansiblePlaybookExeName;
    private String ansiblePlaybookContent;

    public void setupAnsible() {
        GuiLogicTestConfig.AnsibleTest config = getTestConfig().getAnsibleTest();
        if (config.isEnabled()) {
            //isValid of the config checks a lot already, so we can assume a lot here

            ansiblePlaybookExeName = getTestConfig().getAnsibleTest().getPlaybookExe().trim();

            if (config.getPlaybookContent() != null) {
                ansiblePlaybookContent = config.getPlaybookContent();
            } else {
                if (config.getPlaybookUrl() == null)
                    throw new IllegalStateException("Both config.getPlaybookUrl() and config.getPlaybookContent() are null");
                boolean isSpecialGit = config.getPlaybookUrl().trim().startsWith("git ");
                boolean isSpecialGithub = config.getPlaybookUrl().trim().startsWith("github ");

                if (isSpecialGit || isSpecialGithub) {
                    fetchPlaybookFromGit(config.getPlaybookUrl().trim(),
                            isSpecialGithub,
                            config.getPlaybookFileNameInArchive());
                } else {
                    URL url;
                    try {
                        //noinspection ConstantConditions
                        url = new URL(config.getPlaybookUrl());
                    } catch (MalformedURLException e) {
                        throw new RuntimeException("Invalid URL specified for RSpec: \""+config.getPlaybookUrl()+"\"", e);
                    }

                    //for now, we assume it's an archive ONLY AND ALWAYS if playbookFileNameInArchive is set
                    boolean downloadArchive = config.getPlaybookFileNameInArchive() != null;

                    if (!downloadArchive) {
                        fetchPlaybookYmlUrl(config, url);
                    } else {
                        fetchPlaybookArchiveUrl(config, url);
                    }
                }
            }

            if (!(ansibleDir.exists() && ansibleDir.isDirectory())) {
                File parentDir = ansibleDir.getParentFile();
                if ((parentDir != null && !ansibleDir.exists()) && parentDir.exists() && parentDir.isDirectory()) {
                    //noinspection ResultOfMethodCallIgnored
                    ansibleDir.mkdir();
                    if (!(ansibleDir.exists() && ansibleDir.isDirectory()))
                        errorFatal("The ansible directory \"" + ansibleDir.getPath()+ "\" does not exist and could be created.");
                } else
                    errorFatal("The ansible directory \"" + ansibleDir.getPath() + "\" does not exist and will not be created because the parent dir doesn't exist either.");
            }
            if (!(ansiblePlaybookLocalFile.exists() && ansiblePlaybookLocalFile.isFile() && ansiblePlaybookLocalFile.canRead())) {
                errorFatal("The ansible playbook file \"" + ansiblePlaybookLocalFile.getPath() + "\" does not exist (or is not readable)");
            }

            note("Preparation for ansible test succesfull");
        }
    }

    private void fetchPlaybookFromGit(String gitSourceSpec, boolean isGithub, String playbookFileName) {
        GitFetcherCommand c = GitFetcherCommand.fromString(gitSourceSpec)
                .orElseThrow(() -> new IllegalArgumentException("gitSourceSpec='"+gitSourceSpec+"' is not a valid git source string"));
        ansibleDir = new GitFetcher().fetch(c);

        ansiblePlaybookLocalFile = new File(ansibleDir, playbookFileName);
    }

    private void fetchPlaybookArchiveUrl(GuiLogicTestConfig.AnsibleTest config, URL url) {
        //handle playbook that is an archive
        File archiveFile;
        File archiveDir;
        try {
            //download first, because we don't know how big it is. Might not fit memory in worst case.
            archiveFile = IOUtils.urlToTmpFile(url);
            archiveDir = archiveFile.getParentFile();
        } catch (IOException e) {
            throw new RuntimeException("Failed to fetch Playbook archive at \"" + config.getPlaybookUrl() + "\"", e);
        }

        try {
            ansibleDir = Files.createTempDirectory("ansibleTest").toFile();
        } catch (IOException e) {
            errorFatal("Could not create temporary dir for ansible files.", e);
        }

        ansiblePlaybookLocalFile = new File(ansibleDir, config.getPlaybookFileNameInArchive());

        boolean foundPlayBook = false;
        try {
            //extract files
            List<ArchiveEntry> archiveEntries = new ArrayList<>();

            CompressorStreamFactory compressorStreamFactory = new CompressorStreamFactory();
            ArchiveStreamFactory archiveStreamFactory = new ArchiveStreamFactory();

            //peek to check if this is a compressed archive, like .tar.gz or .tar.bz2 or .tgz
            boolean isCompressedArchive;
            try (InputStream inputStream = new FileInputStream(archiveFile);
                 BufferedInputStream bufferedInputStream = new BufferedInputStream(inputStream);
                 CompressorInputStream cis = compressorStreamFactory.createCompressorInputStream(bufferedInputStream)
            ) {
                isCompressedArchive = true;
            } catch (CompressorException | IOException e) {
                isCompressedArchive = false;
            }
            LOG.debug("Archive is " + (isCompressedArchive ? "a" : "NOT a") + " compressed archive");

            List<String> foundFilenames = new ArrayList<>();

            try (InputStream inputStream = new FileInputStream(archiveFile);
                 BufferedInputStream bufferedInputStream = new BufferedInputStream(inputStream);
                 ArchiveInputStream archiveInput = isCompressedArchive ?
                         archiveStreamFactory.createArchiveInputStream(
                                 new BufferedInputStream(compressorStreamFactory.createCompressorInputStream(bufferedInputStream))
                         ) :
                         archiveStreamFactory.createArchiveInputStream(bufferedInputStream)
            ) {
                ArchiveEntry archiveEntry = archiveInput.getNextEntry();
                while (archiveEntry != null) {
                    archiveEntries.add(archiveEntry);

                    foundFilenames.add(archiveEntry.getName());
                    if (archiveEntry.getName().equals(config.getPlaybookFileNameInArchive())) {
                        foundPlayBook = true;
                    }

                    //extract file to ansibleDir
                    File extractedFile = new File(ansibleDir, archiveEntry.getName());
                    if (archiveEntry.isDirectory()) {
                        boolean success = extractedFile.mkdirs();
                        if (!success)
                            LOG.warn("Failed to create dir to extract to: '" + extractedFile.getAbsolutePath() + "'");
                    } else {
                        Files.copy(archiveInput, extractedFile.toPath());
                    }

                    archiveEntry = archiveInput.getNextEntry();
                }
                archiveInput.close();
            }
        } catch (IOException | CompressorException | ArchiveException e) {
            throw new RuntimeException("Failed to process Playbook archive at \"" + config.getPlaybookUrl() + "\"", e);
        } finally {
            //delete the archive, we don't need it anymore.
            //noinspection ResultOfMethodCallIgnored
            archiveFile.delete();
        }

        if (!foundPlayBook) {
            throw new RuntimeException("Playbook \"" + config.getPlaybookFileNameInArchive() + "\" was not found in archive at \"" + config.getPlaybookUrl() + "\"");
        }
    }

    private void fetchPlaybookYmlUrl(GuiLogicTestConfig.AnsibleTest config, URL url) {
        try (InputStream is = url.openStream()) {
            ansiblePlaybookContent = IOUtils.streamToString(is, "UTF-8");
            if (ansiblePlaybookContent.trim().isEmpty()) {
                warn("Empty Playbook read from URL \"" + ansiblePlaybookContent + "\"");
            }

            String head = TextUtil.abbreviate(ansiblePlaybookContent, 200);
            note("Downloaded Playbook from \"" + config.getPlaybookUrl() + "\". Playbook size: " + ansiblePlaybookContent.length() + " characters.\n" +
                            "Head: " + head,
                    true/*formatAsCode*/);
        } catch (IOException e) {
            throw new RuntimeException("Failed to fetch Playbook at \""+config.getPlaybookUrl()+"\"", e);
        }

        try {
            ansibleDir = Files.createTempDirectory("ansibleTest").toFile();
        } catch (IOException e) {
            errorFatal("Could not create temporary dir for ansible files.", e);
        }

        ansiblePlaybookLocalFile = new File(ansibleDir, "playbook.yml");

        try {
            FileWriter fw = new FileWriter(ansiblePlaybookLocalFile);
            fw.write(ansiblePlaybookContent);
            fw.close();
        } catch (IOException e) {
            errorFatal("Error writing '" + ansiblePlaybookLocalFile.getPath() + "': " + e.getMessage(), e);
        }
    }

    static void checkAnsibleOutput(String output,
                                   ApiTest test,
                                   @Nullable String ansibleSuccessWord, @Nullable String ansibleWarningWord, @Nullable String ansibleFailureWord,
                                   @Nullable String ansibleSuccessRegex, @Nullable String ansibleWarningRegex, @Nullable String ansibleFailureRegex) {
        if (ansibleFailureWord != null && output.contains(ansibleFailureWord))
            test.fatalError("Error: Ansible output contained \"" + ansibleFailureWord + "\"");
        if (ansibleFailureRegex != null && Pattern.compile(ansibleFailureRegex).matcher(output).find())
            test.fatalError("Error: Ansible output matched regex \"" + ansibleFailureRegex + "\"");
        if (ansibleWarningWord != null && output.contains(ansibleWarningWord))
            test.warn("Warning: Ansible output contained \"" + ansibleWarningWord + "\"");
        else {
            if (ansibleWarningRegex != null && Pattern.compile(ansibleWarningRegex).matcher(output).find())
                test.warn("Warning: Ansible output matched regex \"" + ansibleWarningRegex + "\"");
            else {
                if (ansibleSuccessWord != null && output.contains(ansibleSuccessWord))
                    test.note("Success: Ansible output contained \"" + ansibleSuccessWord + "\"");
                else {
                    if (ansibleSuccessRegex != null && Pattern.compile(ansibleSuccessRegex).matcher(output).find())
                        test.note("Success: Ansible output matched regex \"" + ansibleSuccessRegex + "\"");
                    else {
                        if (ansibleSuccessWord != null) {
                            if (ansibleSuccessRegex != null)
                                test.fatalError("Error: Ansible output did not contain \"" + ansibleSuccessWord + "\" or match regex \"" + ansibleSuccessRegex + "\"");
                            else
                                test.fatalError("Error: Ansible output did not contain \"" + ansibleSuccessWord + "\"");
                        } else
                            test.fatalError("Error: Ansible output did not match regex \"" + ansibleSuccessRegex + "\"");
                    }
                }
            }
        }
    }

    private void writeAnsibleFiles(boolean addDebug) throws IOException {
        RspecFactory rspecFactory = RspecFactoryFactory.getRspecFactoryInstance(ModelRspecType.BASIC);
        ResourceInfo resourceInfo = resourceInfos.get(0);
        assertNotNull(resourceInfo.requestRspecSource, "No Request Rspec for resource 0");
        ModelRspec first = resourceInfo.requestRspecSource.getModelRspec(ModelRspecType.BASIC);
        if (first == null) {
            fatalError("Failed to parse request rspec 0");
        }
        ModelRspec mergedRequestRspec = rspecFactory.copyModelRspec(first, "request");

        for (int i = 1; i < resourceInfos.size(); i++) {
            resourceInfo = resourceInfos.get(i);
            assertNotNull(resourceInfo.requestRspecSource, "No Request Rspec for resource " + i);

            ModelRspec cur = resourceInfo.requestRspecSource.getModelRspec(ModelRspecType.BASIC);
            if (cur == null) {
                fatalError("Failed to parse request rspec " + i);
            }

            //copy all nodes to the merged RSpec. (Don't care about links and interfaces.)
            for (RspecNode node : cur.getNodes()) {
                RspecNode mergedNode = rspecFactory.copyNode(mergedRequestRspec, node, RspecNode.InterfaceCopyMethod.NO_COPY);
                mergedRequestRspec.addNode(mergedNode);
                if (addDebug) {
                    note("Added node to merged request RSpec: "+node.getClientId());
                }
            }
        }

        String mergedRequestRspecString = mergedRequestRspec.toGeni3Rspec();
        if (addDebug) {
            currentTestResult.addExtraResult(
                    new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                            .setKey("ansible", "mergedRequestRspec")
                            .setValue(mergedRequestRspecString)
                            .setLarge(true)
                            .setHttpMediaTypeToXml()
            );
        }

        //this uses the alternate manifests of AnsibleFileWriter to get all possible logins from all received manifests

        //safe to copy private key, as it is generated just for this test
        AnsibleFileWriter ansibleFileWriter =
                AnsibleFileWriter.createWithCopiedPrivateKey(
                        new BasicStringRspec(mergedRequestRspecString),
                        nodeLoginTestStep.getSshKeyHelper().getSshPrivateKey(),
                        nodeLoginTestStep.getSshKeyHelper().getSshPublicKey(),
                        getUser(),
                        null,
                        nodeLoginTestStep.getSshUsername()
                );


        for (ResourceInfo curResourceInfo : resourceInfos) {
            assert curResourceInfo.manifestRspecString != null;
            if (curResourceInfo.manifestRspecString != null) {
                ansibleFileWriter.addAltBasicStringRspec(curResourceInfo.manifestRspecString);
            }
        }

        ansibleFileWriter.writeFilesToDir(ansibleDir);
        if (addDebug) {
            currentTestResult.addExtraResult(
                    new ApiTestResult.ApiTestMethodResult.FedmonResultExtraBuilder()
                            .setKey("ansible", "hostsFile")
                            .setValue(ansibleFileWriter.getAnsibleHostFileContent())
                            .setLarge(true)
                            .setHttpMediaTypeToText()
            );
        }
    }

    private String callAnsible() {
        ArrayList<String> command = new ArrayList<>();
        command.add(ansiblePlaybookExeName);
        command.add(ansiblePlaybookLocalFile.getAbsolutePath());
        if (getTestConfig().getAnsibleTest().getDebug())
            command.add("-vvvv");
        try {
            ProcessBuilder pb = new ProcessBuilder(command);
            pb.directory(ansibleDir);
            pb = pb.redirectErrorStream(true); //combine stderr and stout, so we can read both at once in one thread. Makes everything easy.

            Process p = pb.start();
            BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));

            String output = "";
            String line;
            while ((line = input.readLine()) != null) {
                output += line + System.lineSeparator();
            }
            //note: input might close before terminal stops.
            input.close();

            try {
                p.waitFor();
            } catch (InterruptedException e) {
                //ignore
            }

            return output;
        } catch (Exception e) {
            fatalError("Error trying to call ansible with command: " + command + "", e);
            return null;
        }
    }

    private class ESpecLogToTest implements ESpecLogListener {

        private final String index;

        private ESpecLogToTest(String index) {
            this.index = index;
        }

        @Override
        public void onPreFileLoad() {
            note("ExperimentSpecification "+index+": Start loading files.");
        }

        @Override
        public void onPostFileLoad(@Nonnull FileSource fileSource, boolean isFast, long byteSize, boolean success) {
            if (!success)
                errorNonFatal("ExperimentSpecification "+index+": Failed to load file \""+fileSource.getBasename()+"\"");
            else {
                if (!isFast)
                    note("ExperimentSpecification "+index+": Successfully loaded file \""+fileSource.getBasename()+"\" ("+byteSize+" byte)");
            }
        }

        @Override
        public void onPostFileLoadAll(boolean success) {
            if (!success)
                errorNonFatal("ExperimentSpecification "+index+": Failed to load all files.");
            else
                note("ExperimentSpecification "+index+": Successfully finished loading files.");
        }

        @Override
        public void onPreRSpec() {
            note("ExperimentSpecification "+index+": Start Provisioning RSpec.");
        }

        @Override
        public void onPostRSpec(@Nonnull RspecSpec rspec, boolean success) {

        }

        @Override
        public void onPostRspecAll(boolean success, @Nonnull List<String> allNodeClientIds) {
            if (!success)
                errorNonFatal("ExperimentSpecification "+index+": Failed to Provision RSpec");
            else
                note("ExperimentSpecification "+index+": Succesfully Provisioned RSpec with nodes "+allNodeClientIds.stream().collect(Collectors.joining()));
        }

        @Override
        public void onPreDir() {
            note("ExperimentSpecification "+index+": Start dir setup.");
        }

        @Override
        public void onPostDir(@Nonnull DirSpec dir, @Nonnull String fullPath, @Nonnull String nodeClientId, boolean success) {
            if (!success)
                errorNonFatal("ExperimentSpecification "+index+": (on \"" + nodeClientId + "\") Failed to setup dir at \""+fullPath+"\"");
            else
                note("ExperimentSpecification "+index+": (on \"" + nodeClientId + "\") Succesfully setup dir at \""+fullPath+"\"");
        }

        @Override
        public void onPostDirAll(boolean success) {
            if (!success)
                errorNonFatal("ExperimentSpecification "+index+": Failed to setup dirs");
            else
                note("ExperimentSpecification "+index+": Succesfully setup dirs");
        }

        @Override
        public void onPreUpload() {
            note("ExperimentSpecification "+index+": Start upload(s).");
        }

        @Override
        public void onPostUpload(@Nonnull UploadLikeSpec uploadSpec, @Nonnull String fullPath, @Nonnull String nodeClientId, boolean success) {
            if (!success)
                errorNonFatal("ExperimentSpecification "+index+": (on \"" + nodeClientId + "\") Failed to upload file \""+uploadSpec.getDesc()+"\" at \""+fullPath+"\"");
            else
                note("ExperimentSpecification "+index+": (on \"" + nodeClientId + "\") Succesfully uploaded file \"" + uploadSpec.getDesc() + "\" at \""+fullPath+"\"");
        }

        @Override
        public void onPostUploadAll(boolean success) {
            if (!success)
                errorNonFatal("ExperimentSpecification "+index+": Failed to upload at least one file to one node");
            else
                note("ExperimentSpecification "+index+": Succesfully uploaded files");
        }

        @Override
        public void onPreExecute() {
            note("ExperimentSpecification "+index+": Start execute(s).");
        }

        @Override
        public void onPostExecute(@Nonnull ExecuteSpec executeSpec, @Nonnull String fullPath, @Nonnull String nodeClientId, int exitStatus, boolean success) {
            if (!success)
                errorNonFatal("ExperimentSpecification "+index+": (on \"" + nodeClientId + "\") Failed to execute script \"" + executeSpec.getDesc() + "\" at \""+fullPath+"\" exitStatus="+exitStatus);
            else
                note("ExperimentSpecification "+index+": (on \"" + nodeClientId + "\") Succesfully executed script \"" + executeSpec.getDesc() + "\" at \""+fullPath+"\" exitStatus="+exitStatus);
        }

        @Override
        public void onPostExecuteStepAll(@Nonnull ExecuteSpec executeSpec, boolean success) {
            if (!success)
                errorNonFatal("ExperimentSpecification "+index+": Failed to execute script \"" + executeSpec.getDesc() + "\" on at least one node");
            else
                note("ExperimentSpecification "+index+": Succesfully executed script \"" + executeSpec.getDesc() + "\" on all nodes");
        }

        @Override
        public void onPostExecuteAll(boolean success) {
            if (!success)
                errorNonFatal("ExperimentSpecification "+index+": Failed to execute at least one script on at least one node");
            else
                note("ExperimentSpecification "+index+": Succesfully executed all scripts");
        }

        @Override
        public void onPreAnsible() {
            note("ExperimentSpecification "+index+": Start ansible(s).");
        }

        @Override
        public void onPostAnsiblePlaybook(@Nonnull AnsiblePlaybookSpec ansiblePlaybookSpec, @Nonnull String fullPath, @Nonnull String nodeClientId, boolean success) {
            if (!success)
                errorNonFatal("ExperimentSpecification "+index+": (on \"" + nodeClientId + "\") Failed to execute ansible playbook \"" + ansiblePlaybookSpec.getDesc() + "\" at \""+fullPath+"\"");
            else
                note("ExperimentSpecification "+index+": (on \"" + nodeClientId + "\") Succesfully executed ansible playbook \"" + ansiblePlaybookSpec.getDesc() + "\" at \""+fullPath+"\"");
        }

        @Override
        public void onPostAnsibleAll(boolean success) {
            if (!success)
                errorNonFatal("ExperimentSpecification "+index+": Failed to ansible at least one script on at least one node");
            else
                note("ExperimentSpecification "+index+": Succesfully ansibled all scripts");
        }
    }
}
